@@ -31,11 +31,13 @@
 #include <limits.h>
 #include <dirent.h>
 #include <stdbool.h>
+#include <ctype.h>
 #include <uuid/uuid.h>
 #include "libbtrfsutil/btrfsutil.h"
 #include "kernel-lib/list.h"
 #include "kernel-lib/sizes.h"
 #include "kernel-lib/list_sort.h"
+#include "kernel-lib/overflow.h"
 #include "kernel-shared/ctree.h"
 #include "kernel-shared/compression.h"
 #include "kernel-shared/volumes.h"
@@ -52,6 +54,8 @@
 #include "common/device-utils.h"
 #include "common/open-utils.h"
 #include "common/parse-utils.h"
+#include "common/sysfs-utils.h"
+#include "common/string-utils.h"
 #include "common/filesystem-utils.h"
 #include "common/format-output.h"
 #include "cmds/commands.h"
@@ -79,6 +83,41 @@ static const char * const cmd_filesystem_df_usage[] = {
 	NULL
 };
 
+static void print_df_by_type(int fd, unsigned int unit_mode) {
+	static const char *files[] = {
+		"bg_reclaim_threshold",
+		"bytes_may_use",
+		"bytes_pinned",
+		"bytes_readonly",
+		"bytes_reserved",
+		"bytes_used",
+		"bytes_zone_unusable",
+		"chunk_size",
+		"disk_total",
+		"disk_used",
+		"total_bytes",
+	};
+	char path[PATH_MAX] = { 0 };
+	const char *types[] = { "data", "metadata", "mixed", "system" };
+	u64 tmp;
+	int ret;
+
+	for (int ti = 0; ti < ARRAY_SIZE(types); ti++) {
+		for (int i = 0; i < ARRAY_SIZE(files); i++) {
+			path_cat3_out(path, "allocation", types[ti], files[i]);
+			ret = sysfs_read_fsid_file_u64(fd, path, &tmp);
+			if (ret < 0)
+				continue;
+			if (i == 0)
+				pr_verbose(LOG_INFO, "%c%s:\n", toupper(types[ti][0]), types[ti] + 1);
+			if (strcmp(files[i], "bg_reclaim_threshold") == 0)
+				pr_verbose(LOG_INFO, "  %-24s  %14llu%%\n", files[i], tmp);
+			else
+				pr_verbose(LOG_INFO, "  %-24s %16s\n", files[i], pretty_size_mode(tmp, unit_mode));
+		}
+	}
+}
+
 static void print_df_text(int fd, struct btrfs_ioctl_space_args *sargs, unsigned unit_mode)
 {
 	u64 i;
@@ -98,6 +137,7 @@ static void print_df_text(int fd, struct btrfs_ioctl_space_args *sargs, unsigned
 			(ok ? ", zone_unusable=" : ""),
 			(ok ? pretty_size_mode(unusable, unit_mode) : ""));
 	}
+	print_df_by_type(fd, unit_mode);
 }
 
 static const struct rowspec filesystem_df_rowspec[] = {
@@ -145,7 +185,6 @@ static int cmd_filesystem_df(const struct cmd_struct *cmd,
 	int ret;
 	int fd;
 	char *path;
-	DIR *dirstream = NULL;
 	unsigned unit_mode;
 
 	unit_mode = get_unit_mode_from_arg(&argc, argv, 1);
@@ -157,7 +196,7 @@ static int cmd_filesystem_df(const struct cmd_struct *cmd,
 
 	path = argv[optind];
 
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0)
 		return 1;
 
@@ -175,7 +214,7 @@ static int cmd_filesystem_df(const struct cmd_struct *cmd,
 	}
 
 	btrfs_warn_multiple_profiles(fd);
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	return !!ret;
 }
 static DEFINE_COMMAND_WITH_FLAGS(filesystem_df, "df", CMD_FORMAT_JSON);
@@ -306,7 +345,7 @@ static void print_one_uuid(struct btrfs_fs_devices *fs_devices,
 	u64 devs_found = 0;
 	u64 total;
 
-	if (add_seen_fsid(fs_devices->fsid, seen_fsid_hash, -1, NULL))
+	if (add_seen_fsid(fs_devices->fsid, seen_fsid_hash, -1))
 		return;
 
 	uuid_unparse(fs_devices->fsid, uuidbuf);
@@ -351,7 +390,7 @@ static int print_one_fs(struct btrfs_ioctl_fs_info_args *fs_info,
 	struct btrfs_ioctl_dev_info_args *tmp_dev_info;
 	int ret;
 
-	ret = add_seen_fsid(fs_info->fsid, seen_fsid_hash, -1, NULL);
+	ret = add_seen_fsid(fs_info->fsid, seen_fsid_hash, -1);
 	if (ret == -EEXIST)
 		return 0;
 	else if (ret)
@@ -707,6 +746,7 @@ static int cmd_filesystem_show(const struct cmd_struct *cmd,
 	struct btrfs_fs_devices *fs_devices;
 	struct btrfs_root *root = NULL;
 	char *search = NULL;
+	char *canon_path = NULL;
 	int ret;
 	/* default, search both kernel and udev */
 	int where = -1;
@@ -788,8 +828,15 @@ static int cmd_filesystem_show(const struct cmd_struct *cmd,
 		}
 	}
 
-	if (where == BTRFS_SCAN_LBLKID)
+	if (where == BTRFS_SCAN_LBLKID) {
+		/*
+		 * Blkid needs canonicalized paths, eg. when the /dev/dm-0 is
+		 * passed on command line.
+		 */
+		canon_path = path_canonicalize(search);
+		search = canon_path;
 		goto devs_only;
+	}
 
 	/* show mounted btrfs */
 	ret = btrfs_scan_kernel(search, unit_mode);
@@ -848,6 +895,7 @@ devs_only:
 		free_fs_devices(fs_devices);
 	}
 out:
+	free(canon_path);
 	if (root)
 		close_ctree(root);
 	free_seen_fsid(seen_fsid_hash);
@@ -903,6 +951,7 @@ static const char * const cmd_filesystem_defrag_usage[] = {
 	OPTLINE("-s start", "defragment only from byte onward"),
 	OPTLINE("-l len", "defragment only up to len bytes"),
 	OPTLINE("-t size", "target extent size hint (default: 32M)"),
+	OPTLINE("--step SIZE", "process the range in given steps, flush after each one"),
 	OPTLINE("-v", "deprecated, alias for global -v option"),
 	HELPINFO_INSERT_GLOBALS,
 	HELPINFO_INSERT_VERBOSE,
@@ -916,6 +965,48 @@ static const char * const cmd_filesystem_defrag_usage[] = {
 
 static struct btrfs_ioctl_defrag_range_args defrag_global_range;
 static int defrag_global_errors;
+static u64 defrag_global_step;
+
+static int defrag_range_in_steps(int fd, const struct stat *st) {
+	int ret = 0;
+	u64 end;
+	struct btrfs_ioctl_defrag_range_args range;
+
+	if (defrag_global_step == 0)
+		return ioctl(fd, BTRFS_IOC_DEFRAG_RANGE, &defrag_global_range);
+
+	/*
+	 * If start is set but length is not within or beyond the u64 range,
+	 * assume it's the rest of the range.
+	 */
+	if (check_add_overflow(defrag_global_range.start, defrag_global_range.len, &end))
+	    end = (u64)-1;
+
+	range = defrag_global_range;
+	range.flags |= BTRFS_DEFRAG_RANGE_START_IO;
+	while (range.start < end) {
+		u64 start;
+
+		range.len = defrag_global_step;
+		pr_verbose(LOG_VERBOSE, "defrag range step: start=%llu len=%llu step=%llu\n",
+			   range.start, range.len, defrag_global_step);
+		ret = ioctl(fd, BTRFS_IOC_DEFRAG_RANGE, &range);
+		if (ret < 0)
+			return ret;
+		if (check_add_overflow(range.start, defrag_global_step, &start))
+			break;
+		range.start = start;
+		/*
+		 * Avoid -EINVAL when starting the next ioctl, this can still
+		 * happen if the file size changes since the time of stat().
+		 */
+		if (start >= (u64)st->st_size)
+			break;
+	}
+
+	return ret;
+}
+
 static int defrag_callback(const char *fpath, const struct stat *sb,
 		int typeflag, struct FTW *ftwbuf)
 {
@@ -928,7 +1019,7 @@ static int defrag_callback(const char *fpath, const struct stat *sb,
 		if (fd < 0) {
 			goto error;
 		}
-		ret = ioctl(fd, BTRFS_IOC_DEFRAG_RANGE, &defrag_global_range);
+		ret = defrag_range_in_steps(fd, sb);
 		close(fd);
 		if (ret && errno == ENOTTY) {
 			error(
@@ -960,7 +1051,6 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 	bool recursive = false;
 	int ret = 0;
 	int compress_type = BTRFS_COMPRESS_NONE;
-	DIR *dirstream;
 
 	/*
 	 * Kernel 4.19+ supports defragmention of files open read-only,
@@ -994,7 +1084,14 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 	defrag_global_errors = 0;
 	optind = 0;
 	while(1) {
-		int c = getopt(argc, argv, "vrc::fs:l:t:");
+		enum { GETOPT_VAL_STEP = GETOPT_VAL_FIRST };
+		static const struct option long_options[] = {
+			{ "step", required_argument, NULL, GETOPT_VAL_STEP },
+			{ NULL, 0, NULL, 0 }
+		};
+		int c;
+
+		c = getopt_long(argc, argv, "vrc::fs:l:t:", long_options, NULL);
 		if (c < 0)
 			break;
 
@@ -1014,13 +1111,13 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 				bconf_be_verbose();
 			break;
 		case 's':
-			start = parse_size_from_string(optarg);
+			start = arg_strtou64_with_suffix(optarg);
 			break;
 		case 'l':
-			len = parse_size_from_string(optarg);
+			len = arg_strtou64_with_suffix(optarg);
 			break;
 		case 't':
-			thresh = parse_size_from_string(optarg);
+			thresh = arg_strtou64_with_suffix(optarg);
 			if (thresh > (u32)-1) {
 				warning(
 			    "target extent size %llu too big, trimmed to %u",
@@ -1031,6 +1128,14 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 		case 'r':
 			recursive = true;
 			break;
+		case GETOPT_VAL_STEP:
+			defrag_global_step = arg_strtou64_with_suffix(optarg);
+			if (defrag_global_step < SZ_256K) {
+				warning("step %llu too small, adjusting to 256KiB\n",
+					   defrag_global_step);
+				defrag_global_step = SZ_256K;
+			}
+			break;
 		default:
 			usage_unknown_option(cmd, argv);
 		}
@@ -1083,11 +1188,9 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 		struct stat st;
 		int defrag_err = 0;
 
-		dirstream = NULL;
-		fd = open_file_or_dir3(argv[i], &dirstream, defrag_open_mode);
+		fd = btrfs_open_path(argv[i], defrag_open_mode == O_RDWR, false);
 		if (fd < 0) {
-			error("cannot open %s: %m", argv[i]);
-			ret = -errno;
+			ret = fd;
 			goto next;
 		}
 
@@ -1112,14 +1215,13 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 			ret = 0;
 		} else {
 			pr_verbose(LOG_INFO, "%s\n", argv[i]);
-			ret = ioctl(fd, BTRFS_IOC_DEFRAG_RANGE,
-					&defrag_global_range);
+			ret = defrag_range_in_steps(fd, &st);
 			defrag_err = errno;
 			if (ret && defrag_err == ENOTTY) {
 				error(
 "defrag range ioctl not supported in this kernel version, 2.6.33 and newer is required");
 				defrag_global_errors++;
-				close_file_or_dir(fd, dirstream);
+				close(fd);
 				break;
 			}
 			if (ret) {
@@ -1131,7 +1233,7 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 next:
 		if (ret)
 			defrag_global_errors++;
-		close_file_or_dir(fd, dirstream);
+		close(fd);
 	}
 
 	if (defrag_global_errors)
@@ -1152,19 +1254,21 @@ static const char * const cmd_filesystem_resize_usage[] = {
 	NULL
 };
 
-static int check_resize_args(const char *amount, const char *path) {
+static int check_resize_args(const char *amount, const char *path, u64 *devid_ret) {
 	struct btrfs_ioctl_fs_info_args fi_args;
 	struct btrfs_ioctl_dev_info_args *di_args = NULL;
 	int ret, i, dev_idx = -1;
 	u64 devid = 1;
+	u64 mindev = (u64)-1;
+	int mindev_idx = 0;
 	const char *res_str = NULL;
 	char *devstr = NULL, *sizestr = NULL;
 	u64 new_size = 0, old_size = 0, diff = 0;
 	int mod = 0;
 	char amount_dup[BTRFS_VOL_NAME_MAX];
 
+	*devid_ret = (u64)-1;
 	ret = get_fs_info(path, &fi_args, &di_args);
-
 	if (ret) {
 		error("unable to retrieve fs info");
 		return 1;
@@ -1182,6 +1286,14 @@ static int check_resize_args(const char *amount, const char *path) {
 		ret = 1;
 		goto out;
 	}
+	ret = 0;
+
+	/* Cancel does not need to determine the device number. */
+	if (strcmp(amount, "cancel") == 0) {
+		/* Different format, print and exit */
+		pr_verbose(LOG_DEFAULT, "Request to cancel resize\n");
+		goto out;
+	}
 
 	sizestr = amount_dup;
 	devstr = strchr(sizestr, ':');
@@ -1202,24 +1314,40 @@ static int check_resize_args(const char *amount, const char *path) {
 
 	dev_idx = -1;
 	for(i = 0; i < fi_args.num_devices; i++) {
+		if (di_args[i].devid < mindev) {
+			mindev = di_args[i].devid;
+			mindev_idx = i;
+		}
 		if (di_args[i].devid == devid) {
 			dev_idx = i;
 			break;
 		}
 	}
 
-	if (dev_idx < 0) {
+	if (devstr && dev_idx < 0) {
+		/* Devid specified but not found. */
 		error("cannot find devid: %lld", devid);
 		ret = 1;
 		goto out;
+	} else if (!devstr && devid == 1 && dev_idx < 0) {
+		/*
+		 * No device specified, assuming implicit 1 but it does not
+		 * exist. Use minimum device as fallback.
+		 */
+		warning("no devid specified means devid 1 which does not exist, using\n"
+			"\t lowest devid %llu as a fallback", mindev);
+		*devid_ret = mindev;
+		devid = mindev;
+		dev_idx = mindev_idx;
+	} else {
+		/*
+		 * Use the initial value 1 or the parsed number but don't
+		 * return it by devid_ret as the resize string works as-is.
+		 */
 	}
 
 	if (strcmp(sizestr, "max") == 0) {
 		res_str = "max";
-	} else if (strcmp(sizestr, "cancel") == 0) {
-		/* Different format, print and exit */
-		pr_verbose(LOG_DEFAULT, "Request to cancel resize\n");
-		goto out;
 	} else {
 		if (sizestr[0] == '-') {
 			mod = -1;
@@ -1228,8 +1356,8 @@ static int check_resize_args(const char *amount, const char *path) {
 			mod = 1;
 			sizestr++;
 		}
-		diff = parse_size_from_string(sizestr);
-		if (!diff) {
+		ret = parse_u64_with_suffix(sizestr, &diff);
+		if (ret < 0) {
 			error("failed to parse size %s", sizestr);
 			ret = 1;
 			goto out;
@@ -1268,7 +1396,7 @@ static int check_resize_args(const char *amount, const char *path) {
 
 out:
 	free(di_args);
-	return 0;
+	return ret;
 }
 
 static int cmd_filesystem_resize(const struct cmd_struct *cmd,
@@ -1277,7 +1405,7 @@ static int cmd_filesystem_resize(const struct cmd_struct *cmd,
 	struct btrfs_ioctl_vol_args	args;
 	int	fd, res, len, e;
 	char	*amount, *path;
-	DIR	*dirstream = NULL;
+	u64 devid;
 	int ret;
 	bool enqueue = false;
 	bool cancel = false;
@@ -1314,10 +1442,10 @@ static int cmd_filesystem_resize(const struct cmd_struct *cmd,
 
 	cancel = (strcmp("cancel", amount) == 0);
 
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0) {
 		/* The path is a directory */
-		if (fd == -3) {
+		if (fd == -ENOTDIR) {
 			error(
 		"resize works on mounted filesystems and accepts only\n"
 		"directories as argument. Passing file containing a btrfs image\n"
@@ -1337,22 +1465,29 @@ static int cmd_filesystem_resize(const struct cmd_struct *cmd,
 			if (ret < 0)
 				error(
 			"unable to check status of exclusive operation: %m");
-			close_file_or_dir(fd, dirstream);
+			close(fd);
 			return 1;
 		}
 	}
 
-	ret = check_resize_args(amount, path);
+	ret = check_resize_args(amount, path, &devid);
 	if (ret != 0) {
-		close_file_or_dir(fd, dirstream);
+		close(fd);
 		return 1;
 	}
 
 	memset(&args, 0, sizeof(args));
-	strncpy_null(args.name, amount);
+	if (devid == (u64)-1) {
+		/* Ok to copy the string verbatim. */
+		strncpy_null(args.name, amount, sizeof(args.name));
+	} else {
+		/* The implicit devid 1 needs to be adjusted. */
+		snprintf(args.name, sizeof(args.name) - 1, "%llu:%s", devid, amount);
+	}
+	pr_verbose(LOG_VERBOSE, "adjust resize argument to: %s\n", args.name);
 	res = ioctl(fd, BTRFS_IOC_RESIZE, &args);
 	e = errno;
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	if( res < 0 ){
 		switch (e) {
 		case EFBIG:
@@ -1513,7 +1648,7 @@ static int cmd_filesystem_mkswapfile(const struct cmd_struct *cmd, int argc, cha
 
 		switch (c) {
 		case 's':
-			size = parse_size_from_string(optarg);
+			size = arg_strtou64_with_suffix(optarg);
 			/* Minimum limit reported by mkswap */
 			if (size < 40 * SZ_1K) {
 				error("swapfile needs to be at least 40 KiB");
