@@ -20,8 +20,19 @@
 #define __BTRFS_DISK_IO_H__
 
 #include "kerncompat.h"
-#include "kernel-shared/ctree.h"
+#include <stddef.h>
+#include "kernel-lib/bitops.h"
 #include "kernel-lib/sizes.h"
+#include "kernel-shared/ctree.h"
+#include "kernel-shared/uapi/btrfs.h"
+
+struct btrfs_tree_parent_check;
+struct btrfs_fs_devices;
+struct btrfs_key;
+struct btrfs_super_block;
+struct btrfs_trans_handle;
+struct extent_buffer;
+struct rb_node;
 
 #define BTRFS_SUPER_MIRROR_MAX	 3
 #define BTRFS_SUPER_MIRROR_SHIFT 12
@@ -117,19 +128,19 @@ enum btrfs_open_ctree_flags {
 enum btrfs_read_sb_flags {
 	SBREAD_DEFAULT		= 0,
 	/* Reading superblock during recovery */
-	SBREAD_RECOVER		= (1 << 0),
+	SBREAD_RECOVER		= (1U << 0),
 
 	/*
 	 * Read superblock with the fake signature, cannot be used with
 	 * SBREAD_RECOVER
 	 */
-	SBREAD_TEMPORARY = (1 << 1),
+	SBREAD_TEMPORARY	= (1U << 1),
 
 	/*
 	 * Equivalent of OPEN_CTREE_IGNORE_FSID_MISMATCH, allow to read
 	 * superblock that has mismatched sb::fsid and sb::dev_item.fsid
 	 */
-	SBREAD_IGNORE_FSID_MISMATCH = (1 << 2),
+	SBREAD_IGNORE_FSID_MISMATCH = (1U << 2),
 };
 
 /*
@@ -150,8 +161,7 @@ struct btrfs_device;
 
 int read_whole_eb(struct btrfs_fs_info *info, struct extent_buffer *eb, int mirror);
 struct extent_buffer *read_tree_block(struct btrfs_fs_info *fs_info, u64 bytenr,
-				      u64 owner_root, u64 parent_transid,
-				      int level, struct btrfs_key *first_key);
+				      struct btrfs_tree_parent_check *check);
 
 void readahead_tree_block(struct btrfs_fs_info *fs_info, u64 bytenr,
 			  u64 parent_transid);
@@ -227,9 +237,8 @@ int btrfs_read_buffer(struct extent_buffer *buf, u64 parent_transid);
 int write_tree_block(struct btrfs_trans_handle *trans,
 		     struct btrfs_fs_info *fs_info,
 		     struct extent_buffer *eb);
-int btrfs_fs_roots_compare_roots(struct rb_node *node1, struct rb_node *node2);
+int btrfs_fs_roots_compare_roots(const struct rb_node *node1, const struct rb_node *node2);
 struct btrfs_root *btrfs_create_tree(struct btrfs_trans_handle *trans,
-				     struct btrfs_fs_info *fs_info,
 				     struct btrfs_key *key);
 int btrfs_delete_and_free_root(struct btrfs_trans_handle *trans,
 			       struct btrfs_root *root);
@@ -243,8 +252,8 @@ int btrfs_global_root_insert(struct btrfs_fs_info *fs_info,
 int btrfs_find_and_setup_root(struct btrfs_root *tree_root,
 			      struct btrfs_fs_info *fs_info,
 			      u64 objectid, struct btrfs_root *root);
-int btrfs_read_extent_buffer(struct extent_buffer *eb, u64 parent_transid,
-			     int level, struct btrfs_key *first_key);
+int btrfs_read_extent_buffer(struct extent_buffer *eb,
+			     struct btrfs_tree_parent_check *check);
 
 static inline struct btrfs_root *btrfs_block_group_root(
 						struct btrfs_fs_info *fs_info)
