@@ -644,10 +644,10 @@ again:
 		device->writeable = 0;
 		list_del(&device->dev_list);
 		/* free the memory */
-		free(device->name);
-		free(device->label);
-		free(device->zone_info);
-		free(device);
+		kfree(device->name);
+		kfree(device->label);
+		kfree(device->zone_info);
+		kfree(device);
 	}
 
 	seed_devices = fs_devices->seed;
@@ -658,11 +658,11 @@ again:
 		orig = fs_devices;
 		fs_devices = seed_devices;
 		list_del(&orig->fs_list);
-		free(orig);
+		kfree(orig);
 		goto again;
 	} else {
 		list_del(&fs_devices->fs_list);
-		free(fs_devices);
+		kfree(fs_devices);
 	}
 
 	return ret;
@@ -871,8 +871,8 @@ again:
 	path->reada = READA_FORWARD;
 
 	key.objectid = device->devid;
-	key.offset = search_start;
 	key.type = BTRFS_DEV_EXTENT_KEY;
+	key.offset = search_start;
 
 	ret = btrfs_search_slot(NULL, root, &key, path, 0, 0);
 	if (ret < 0)
@@ -1006,8 +1006,8 @@ int btrfs_insert_dev_extent(struct btrfs_trans_handle *trans,
 		return -ENOMEM;
 
 	key.objectid = device->devid;
-	key.offset = start;
 	key.type = BTRFS_DEV_EXTENT_KEY;
+	key.offset = start;
 	ret = btrfs_insert_empty_item(trans, root, path, &key,
 				      sizeof(*extent));
 	if (ret < 0)
@@ -1062,8 +1062,8 @@ static int find_next_chunk(struct btrfs_fs_info *fs_info, u64 *offset)
 		return -ENOMEM;
 
 	key.objectid = BTRFS_FIRST_CHUNK_TREE_OBJECTID;
-	key.offset = (u64)-1;
 	key.type = BTRFS_CHUNK_ITEM_KEY;
+	key.offset = (u64)-1;
 
 	ret = btrfs_search_slot(NULL, root, &key, path, 0, 0);
 	if (ret < 0)
@@ -1302,8 +1302,8 @@ static int btrfs_device_avail_bytes(struct btrfs_trans_handle *trans,
 		return -ENOMEM;
 
 	key.objectid = device->devid;
-	key.offset = search_start;
 	key.type = BTRFS_DEV_EXTENT_KEY;
+	key.offset = search_start;
 
 	path->reada = READA_FORWARD;
 	ret = btrfs_search_slot(trans, root, &key, path, 0, 0);
@@ -2663,8 +2663,8 @@ int btrfs_read_chunk_tree(struct btrfs_fs_info *fs_info)
 	 * item - BTRFS_FIRST_CHUNK_TREE_OBJECTID).
 	 */
 	key.objectid = BTRFS_DEV_ITEMS_OBJECTID;
-	key.offset = 0;
 	key.type = 0;
+	key.offset = 0;
 	ret = btrfs_search_slot(NULL, root, &key, path, 0, 0);
 	if (ret < 0)
 		goto error;
@@ -2797,12 +2797,12 @@ static int split_eb_for_raid56(struct btrfs_fs_info *info,
 		}
 		ebs[i] = eb;
 	}
-	free(tmp_ebs);
+	kfree(tmp_ebs);
 	return ret;
 clean_up:
 	for (i = 0; i < num_stripes; i++)
-		free(tmp_ebs[i]);
-	free(tmp_ebs);
+		kfree(tmp_ebs[i]);
+	kfree(tmp_ebs);
 	return ret;
 }
 
@@ -2817,11 +2817,11 @@ int write_raid56_with_parity(struct btrfs_fs_info *info,
 	int alloc_size = eb->len;
 	void **pointers;
 
-	ebs = malloc(sizeof(*ebs) * multi->num_stripes);
-	pointers = malloc(sizeof(*pointers) * multi->num_stripes);
+	ebs = kmalloc(sizeof(*ebs) * multi->num_stripes, GFP_KERNEL);
+	pointers = kmalloc(sizeof(*pointers) * multi->num_stripes, GFP_KERNEL);
 	if (!ebs || !pointers) {
-		free(ebs);
-		free(pointers);
+		kfree(ebs);
+		kfree(pointers);
 		return -ENOMEM;
 	}
 
@@ -2842,7 +2842,7 @@ int write_raid56_with_parity(struct btrfs_fs_info *info,
 			}
 			continue;
 		}
-		new_eb = malloc(sizeof(*eb) + alloc_size);
+		new_eb = kmalloc(sizeof(*eb) + alloc_size, GFP_KERNEL);
 		if (!new_eb) {
 			ret = -ENOMEM;
 			goto out_free_split;
@@ -2885,11 +2885,11 @@ int write_raid56_with_parity(struct btrfs_fs_info *info,
 out_free_split:
 	for (i = 0; i < multi->num_stripes; i++) {
 		if (ebs[i] != eb)
-			free(ebs[i]);
+			kfree(ebs[i]);
 	}
 out:
-	free(ebs);
-	free(pointers);
+	kfree(ebs);
+	kfree(pointers);
 
 	return ret;
 }
