@@ -5,7 +5,7 @@
  *
  * libbtrfsutil is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
+ * the Free Software Foundation, either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * libbtrfsutil is distributed in the hope that it will be useful,
@@ -17,16 +17,17 @@
  * along with libbtrfsutil.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <sys/statfs.h>
+#include <sys/ioctl.h>
+#include <sys/stat.h>
+#include <linux/magic.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <sys/ioctl.h>
-#include <sys/stat.h>
-#include <sys/types.h>
-#include <sys/vfs.h>
-#include <linux/magic.h>
+#include <stdbool.h>
+#include <stdint.h>
 #include "stubs.h"
 
 #include "btrfsutil_internal.h"
@@ -69,6 +70,8 @@ PUBLIC enum btrfs_util_error btrfs_util_is_subvolume(const char *path)
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_is_valid(const char *path)
+LIBBTRFSUTIL_ALIAS(btrfs_util_is_subvolume);
 
 PUBLIC enum btrfs_util_error btrfs_util_is_subvolume_fd(int fd)
 {
@@ -96,6 +99,8 @@ PUBLIC enum btrfs_util_error btrfs_util_is_subvolume_fd(int fd)
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_is_valid_fd(int fd)
+LIBBTRFSUTIL_ALIAS(btrfs_util_is_subvolume_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_subvolume_id(const char *path,
 						     uint64_t *id_ret)
@@ -111,6 +116,8 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_id(const char *path,
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_id(const char *path, uint64_t *id_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_id);
 
 PUBLIC enum btrfs_util_error btrfs_util_subvolume_id_fd(int fd,
 							uint64_t *id_ret)
@@ -122,15 +129,15 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_id_fd(int fd,
 	int ret;
 
 	ret = ioctl(fd, BTRFS_IOC_INO_LOOKUP, &args);
-	if (ret == -1) {
-		close(fd);
+	if (ret == -1)
 		return BTRFS_UTIL_ERROR_INO_LOOKUP_FAILED;
-	}
 
 	*id_ret = args.treeid;
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_id_fd(int fd, uint64_t *id_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_id_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_subvolume_path(const char *path,
 						       uint64_t id,
@@ -147,6 +154,8 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_path(const char *path,
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_path(const char *path, uint64_t id, char **path_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_path);
 
 PUBLIC enum btrfs_util_error btrfs_util_subvolume_path_fd(int fd, uint64_t id,
 							  char **path_ret)
@@ -211,12 +220,12 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_path_fd(int fd, uint64_t id,
 		header = (struct btrfs_ioctl_search_header *)search.buf;
 		ref = (struct btrfs_root_ref *)(header + 1);
 		name = (char *)(ref + 1);
-		name_len = le16_to_cpu(ref->name_len);
+		name_len = get_unaligned_le16(&ref->name_len);
 
-		id = header->offset;
+		id = btrfs_search_header_offset(header);
 
 		lookup.treeid = id;
-		lookup.objectid = le64_to_cpu(ref->dirid);
+		lookup.objectid = get_unaligned_le64(&ref->dirid);
 		ret = ioctl(fd, BTRFS_IOC_INO_LOOKUP, &lookup);
 		if (ret == -1) {
 			free(path);
@@ -256,28 +265,30 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_path_fd(int fd, uint64_t id,
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_path_fd(int fd, uint64_t id, char **path_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_path_fd);
 
+/* The @timespec could be from a raw buffer, do not assume any alignment. */
 static void copy_timespec(struct timespec *timespec,
 			  const struct btrfs_timespec *btrfs_timespec)
 {
-	timespec->tv_sec = le64_to_cpu(btrfs_timespec->sec);
-	timespec->tv_nsec = le32_to_cpu(btrfs_timespec->nsec);
+	timespec->tv_sec = get_unaligned_le64(&btrfs_timespec->sec);
+	timespec->tv_nsec = get_unaligned_le32(&btrfs_timespec->nsec);
 }
 
+/* The @root could be from a raw search buffer, do not assume any alignment. */
 static void copy_root_item(struct btrfs_util_subvolume_info *subvol,
 			   const struct btrfs_root_item *root)
 {
-	subvol->flags = le64_to_cpu(root->flags);
+	subvol->flags = get_unaligned_le64(&root->flags);
 	memcpy(subvol->uuid, root->uuid, sizeof(subvol->uuid));
-	memcpy(subvol->parent_uuid, root->parent_uuid,
-	       sizeof(subvol->parent_uuid));
-	memcpy(subvol->received_uuid, root->received_uuid,
-	       sizeof(subvol->received_uuid));
-	subvol->generation = le64_to_cpu(root->generation);
-	subvol->ctransid = le64_to_cpu(root->ctransid);
-	subvol->otransid = le64_to_cpu(root->otransid);
-	subvol->stransid = le64_to_cpu(root->stransid);
-	subvol->rtransid = le64_to_cpu(root->rtransid);
+	memcpy(subvol->parent_uuid, root->parent_uuid, sizeof(subvol->parent_uuid));
+	memcpy(subvol->received_uuid, root->received_uuid, sizeof(subvol->received_uuid));
+	subvol->generation = get_unaligned_le64(&root->generation);
+	subvol->ctransid = get_unaligned_le64(&root->ctransid);
+	subvol->otransid = get_unaligned_le64(&root->otransid);
+	subvol->stransid = get_unaligned_le64(&root->stransid);
+	subvol->rtransid = get_unaligned_le64(&root->rtransid);
 	copy_timespec(&subvol->ctime, &root->ctime);
 	copy_timespec(&subvol->otime, &root->otime);
 	copy_timespec(&subvol->stime, &root->stime);
@@ -299,6 +310,9 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_info(const char *path,
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_info(const char *path, uint64_t id,
+							   struct btrfs_util_subvolume_info *subvol)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_info);
 
 static enum btrfs_util_error get_subvolume_info_privileged(int fd, uint64_t id,
 							   struct btrfs_util_subvolume_info *subvol)
@@ -360,7 +374,7 @@ static enum btrfs_util_error get_subvolume_info_privileged(int fd, uint64_t id,
 		}
 
 		header = (struct btrfs_ioctl_search_header *)(search.buf + buf_off);
-		if (header->type == BTRFS_ROOT_ITEM_KEY) {
+		if (btrfs_search_header_type(header) == BTRFS_ROOT_ITEM_KEY) {
 			if (subvol) {
 				const struct btrfs_root_item *root;
 
@@ -369,20 +383,20 @@ static enum btrfs_util_error get_subvolume_info_privileged(int fd, uint64_t id,
 			}
 			need_root_item = false;
 			search.key.min_type = BTRFS_ROOT_BACKREF_KEY;
-		} else if (header->type == BTRFS_ROOT_BACKREF_KEY) {
+		} else if (btrfs_search_header_type(header) == BTRFS_ROOT_BACKREF_KEY) {
 			if (subvol) {
 				const struct btrfs_root_ref *ref;
 
 				ref = (const struct btrfs_root_ref *)(header + 1);
-				subvol->parent_id = header->offset;
-				subvol->dir_id = le64_to_cpu(ref->dirid);
+				subvol->parent_id = btrfs_search_header_offset(header);
+				subvol->dir_id = get_unaligned_le64(&ref->dirid);
 			}
 			need_root_backref = false;
 			search.key.min_type = UINT32_MAX;
 		}
 
 		items_pos++;
-		buf_off += sizeof(*header) + header->len;
+		buf_off += sizeof(*header) + btrfs_search_header_len(header);
 	}
 
 	return BTRFS_UTIL_OK;
@@ -453,6 +467,9 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_info_fd(int fd, uint64_t id,
 
 	return get_subvolume_info_privileged(fd, id, subvol);
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_info_fd(int fd, uint64_t id,
+							      struct btrfs_util_subvolume_info *subvol)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_info_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_get_subvolume_read_only_fd(int fd,
 								   bool *read_only_ret)
@@ -467,6 +484,8 @@ PUBLIC enum btrfs_util_error btrfs_util_get_subvolume_read_only_fd(int fd,
 	*read_only_ret = flags & BTRFS_SUBVOL_RDONLY;
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_read_only_fd(int fd, bool *ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_get_subvolume_read_only_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_get_subvolume_read_only(const char *path,
 								bool *ret)
@@ -482,6 +501,8 @@ PUBLIC enum btrfs_util_error btrfs_util_get_subvolume_read_only(const char *path
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_read_only(const char *path, bool *ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_get_subvolume_read_only);
 
 PUBLIC enum btrfs_util_error btrfs_util_set_subvolume_read_only(const char *path,
 								bool read_only)
@@ -497,6 +518,8 @@ PUBLIC enum btrfs_util_error btrfs_util_set_subvolume_read_only(const char *path
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_set_read_only(const char *path, bool read_only)
+LIBBTRFSUTIL_ALIAS(btrfs_util_set_subvolume_read_only);
 
 PUBLIC enum btrfs_util_error btrfs_util_set_subvolume_read_only_fd(int fd,
 								   bool read_only)
@@ -519,6 +542,8 @@ PUBLIC enum btrfs_util_error btrfs_util_set_subvolume_read_only_fd(int fd,
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_set_read_only_fd(int fd, bool read_only)
+LIBBTRFSUTIL_ALIAS(btrfs_util_set_subvolume_read_only_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_get_default_subvolume(const char *path,
 							      uint64_t *id_ret)
@@ -534,6 +559,8 @@ PUBLIC enum btrfs_util_error btrfs_util_get_default_subvolume(const char *path,
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_default(const char *path, uint64_t *id_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_get_default_subvolume);
 
 PUBLIC enum btrfs_util_error btrfs_util_get_default_subvolume_fd(int fd,
 								 uint64_t *id_ret)
@@ -573,27 +600,29 @@ PUBLIC enum btrfs_util_error btrfs_util_get_default_subvolume_fd(int fd,
 		}
 
 		header = (struct btrfs_ioctl_search_header *)(search.buf + buf_off);
-		if (header->type == BTRFS_DIR_ITEM_KEY) {
+		if (btrfs_search_header_type(header) == BTRFS_DIR_ITEM_KEY) {
 			const struct btrfs_dir_item *dir;
 			const char *name;
 			uint16_t name_len;
 
 			dir = (struct btrfs_dir_item *)(header + 1);
 			name = (const char *)(dir + 1);
-			name_len = le16_to_cpu(dir->name_len);
+			name_len = get_unaligned_le16(&dir->name_len);
 			if (strncmp(name, "default", name_len) == 0) {
-				*id_ret = le64_to_cpu(dir->location.objectid);
+				*id_ret = get_unaligned_le64(&dir->location.objectid);
 				break;
 			}
 		}
 
 		items_pos++;
-		buf_off += sizeof(*header) + header->len;
-		search.key.min_offset = header->offset + 1;
+		buf_off += sizeof(*header) + btrfs_search_header_len(header);
+		search.key.min_offset = btrfs_search_header_offset(header) + 1;
 	}
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_get_default_fd(int fd, uint64_t *id_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_get_default_subvolume_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_set_default_subvolume(const char *path,
 							      uint64_t id)
@@ -609,6 +638,8 @@ PUBLIC enum btrfs_util_error btrfs_util_set_default_subvolume(const char *path,
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_set_default(const char *path, uint64_t id)
+LIBBTRFSUTIL_ALIAS(btrfs_util_set_default_subvolume);
 
 PUBLIC enum btrfs_util_error btrfs_util_set_default_subvolume_fd(int fd,
 								 uint64_t id)
@@ -632,6 +663,8 @@ PUBLIC enum btrfs_util_error btrfs_util_set_default_subvolume_fd(int fd,
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_set_default_fd(int fd, uint64_t id)
+LIBBTRFSUTIL_ALIAS(btrfs_util_set_default_subvolume_fd);
 
 static enum btrfs_util_error openat_parent_and_name(int dirfd, const char *path,
 						    char *name, size_t name_len,
@@ -685,7 +718,7 @@ static enum btrfs_util_error openat_parent_and_name(int dirfd, const char *path,
 
 PUBLIC enum btrfs_util_error btrfs_util_create_subvolume(const char *path,
 							 int flags,
-							 uint64_t *async_transid,
+							 uint64_t *unused,
 							 struct btrfs_util_qgroup_inherit *qgroup_inherit)
 {
 	char name[BTRFS_SUBVOL_NAME_MAX + 1];
@@ -698,15 +731,19 @@ PUBLIC enum btrfs_util_error btrfs_util_create_subvolume(const char *path,
 		return err;
 
 	err = btrfs_util_create_subvolume_fd(parent_fd, name, flags,
-					    async_transid, qgroup_inherit);
+					    unused, qgroup_inherit);
 	SAVE_ERRNO_AND_CLOSE(parent_fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_create(const char *path, int flags,
+							 uint64_t *unused,
+							 struct btrfs_util_qgroup_inherit *qgroup_inherit)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_subvolume);
 
 PUBLIC enum btrfs_util_error btrfs_util_create_subvolume_fd(int parent_fd,
 							    const char *name,
 							    int flags,
-							    uint64_t *async_transid,
+							    uint64_t *unused,
 							    struct btrfs_util_qgroup_inherit *qgroup_inherit)
 {
 	struct btrfs_ioctl_vol_args_v2 args = {};
@@ -718,8 +755,6 @@ PUBLIC enum btrfs_util_error btrfs_util_create_subvolume_fd(int parent_fd,
 		return BTRFS_UTIL_ERROR_INVALID_ARGUMENT;
 	}
 
-	if (async_transid)
-		args.flags |= BTRFS_SUBVOL_CREATE_ASYNC;
 	if (qgroup_inherit) {
 		args.flags |= BTRFS_SUBVOL_QGROUP_INHERIT;
 		args.qgroup_inherit = (struct btrfs_qgroup_inherit *)qgroup_inherit;
@@ -740,13 +775,16 @@ PUBLIC enum btrfs_util_error btrfs_util_create_subvolume_fd(int parent_fd,
 	if (ret == -1)
 		return BTRFS_UTIL_ERROR_SUBVOL_CREATE_FAILED;
 
-	if (async_transid)
-		*async_transid = args.transid;
-
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_create_fd(int parent_fd,
+							    const char *name,
+							    int flags,
+							    uint64_t *unused,
+							    struct btrfs_util_qgroup_inherit *qgroup_inherit)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_subvolume_fd);
 
-#define BTRFS_UTIL_SUBVOLUME_ITERATOR_CLOSE_FD (1 << 30)
+#define BTRFS_UTIL_SUBVOLUME_ITERATOR_CLOSE_FD (1U << 30)
 
 struct search_stack_entry {
 	union {
@@ -892,7 +930,8 @@ static enum btrfs_util_error append_to_search_stack(struct btrfs_util_subvolume_
 				return err;
 			}
 
-			close(iter->cur_fd);
+			if (iter->cur_fd != iter->fd)
+				close(iter->cur_fd);
 			iter->cur_fd = fd;
 		}
 	}
@@ -955,6 +994,11 @@ PUBLIC enum btrfs_util_error btrfs_util_create_subvolume_iterator(const char *pa
 
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_iter_create(const char *path,
+							      uint64_t top,
+							      int flags,
+							      struct btrfs_util_subvolume_iterator **ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_subvolume_iterator);
 
 PUBLIC enum btrfs_util_error btrfs_util_create_subvolume_iterator_fd(int fd,
 								     uint64_t top,
@@ -1022,10 +1066,14 @@ out_iter:
 	free(iter);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_iter_create_fd(int fd,
+								 uint64_t top,
+								 int flags,
+								 struct btrfs_util_subvolume_iterator **ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_subvolume_iterator_fd);
 
 static enum btrfs_util_error snapshot_subvolume_children(int fd, int parent_fd,
-							 const char *name,
-							 uint64_t *async_transid)
+							 const char *name)
 {
 	struct btrfs_util_subvolume_iterator *iter;
 	enum btrfs_util_error err;
@@ -1043,7 +1091,6 @@ static enum btrfs_util_error snapshot_subvolume_children(int fd, int parent_fd,
 		char child_name[BTRFS_SUBVOL_NAME_MAX + 1];
 		char *child_path;
 		int child_fd, new_parent_fd;
-		uint64_t tmp_transid;
 
 		err = btrfs_util_subvolume_iterator_next(iter, &child_path,
 							 NULL);
@@ -1078,14 +1125,11 @@ static enum btrfs_util_error snapshot_subvolume_children(int fd, int parent_fd,
 
 		err = btrfs_util_create_snapshot_fd2(child_fd, new_parent_fd,
 						     child_name, 0,
-						     async_transid ? &tmp_transid : NULL,
-						     NULL);
+						     NULL, NULL);
 		SAVE_ERRNO_AND_CLOSE(child_fd);
 		SAVE_ERRNO_AND_CLOSE(new_parent_fd);
 		if (err)
 			break;
-		if (async_transid && tmp_transid > *async_transid)
-			*async_transid = tmp_transid;
 	}
 
 	btrfs_util_destroy_subvolume_iterator(iter);
@@ -1097,7 +1141,7 @@ out:
 PUBLIC enum btrfs_util_error btrfs_util_create_snapshot(const char *source,
 							const char *path,
 							int flags,
-							uint64_t *async_transid,
+							uint64_t *unused,
 							struct btrfs_util_qgroup_inherit *qgroup_inherit)
 {
 	enum btrfs_util_error err;
@@ -1107,16 +1151,21 @@ PUBLIC enum btrfs_util_error btrfs_util_create_snapshot(const char *source,
 	if (fd == -1)
 		return BTRFS_UTIL_ERROR_OPEN_FAILED;
 
-	err = btrfs_util_create_snapshot_fd(fd, path, flags, async_transid,
+	err = btrfs_util_create_snapshot_fd(fd, path, flags, unused,
 					    qgroup_inherit);
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_snapshot(const char *source,
+							   const char *path, int flags,
+							   uint64_t *unused,
+							   struct btrfs_util_qgroup_inherit *qgroup_inherit)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_snapshot);
 
 PUBLIC enum btrfs_util_error btrfs_util_create_snapshot_fd(int fd,
 							   const char *path,
 							   int flags,
-							   uint64_t *async_transid,
+							   uint64_t *unused,
 							   struct btrfs_util_qgroup_inherit *qgroup_inherit)
 {
 	char name[BTRFS_SUBVOL_NAME_MAX + 1];
@@ -1129,16 +1178,21 @@ PUBLIC enum btrfs_util_error btrfs_util_create_snapshot_fd(int fd,
 		return err;
 
 	err = btrfs_util_create_snapshot_fd2(fd, parent_fd, name, flags,
-					     async_transid, qgroup_inherit);
+					     unused, qgroup_inherit);
 	SAVE_ERRNO_AND_CLOSE(parent_fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_snapshot_fd(int fd, const char *path,
+							      int flags,
+							      uint64_t *unused,
+							      struct btrfs_util_qgroup_inherit *qgroup_inherit)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_snapshot_fd);
 
 PUBLIC enum btrfs_util_error btrfs_util_create_snapshot_fd2(int fd,
 							    int parent_fd,
 							    const char *name,
 							    int flags,
-							    uint64_t *async_transid,
+							    uint64_t *unused,
 							    struct btrfs_util_qgroup_inherit *qgroup_inherit)
 {
 	struct btrfs_ioctl_vol_args_v2 args = {.fd = fd};
@@ -1155,8 +1209,6 @@ PUBLIC enum btrfs_util_error btrfs_util_create_snapshot_fd2(int fd,
 
 	if (flags & BTRFS_UTIL_CREATE_SNAPSHOT_READ_ONLY)
 		args.flags |= BTRFS_SUBVOL_RDONLY;
-	if (async_transid)
-		args.flags |= BTRFS_SUBVOL_CREATE_ASYNC;
 	if (qgroup_inherit) {
 		args.flags |= BTRFS_SUBVOL_QGROUP_INHERIT;
 		args.qgroup_inherit = (struct btrfs_qgroup_inherit *)qgroup_inherit;
@@ -1177,18 +1229,20 @@ PUBLIC enum btrfs_util_error btrfs_util_create_snapshot_fd2(int fd,
 	if (ret == -1)
 		return BTRFS_UTIL_ERROR_SUBVOL_CREATE_FAILED;
 
-	if (async_transid)
-		*async_transid = args.transid;
-
 	if (flags & BTRFS_UTIL_CREATE_SNAPSHOT_RECURSIVE) {
-		err = snapshot_subvolume_children(fd, parent_fd, name,
-						  async_transid);
+		err = snapshot_subvolume_children(fd, parent_fd, name);
 		if (err)
 			return err;
 	}
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_snapshot_fd2(int fd, int parent_fd,
+							       const char *name,
+							       int flags,
+							       uint64_t *unused,
+							       struct btrfs_util_qgroup_inherit *qgroup_inherit)
+LIBBTRFSUTIL_ALIAS(btrfs_util_create_snapshot_fd2);
 
 static enum btrfs_util_error delete_subvolume_children(int parent_fd,
 						       const char *name)
@@ -1256,6 +1310,8 @@ PUBLIC enum btrfs_util_error btrfs_util_delete_subvolume(const char *path,
 	SAVE_ERRNO_AND_CLOSE(parent_fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_delete(const char *path, int flags)
+LIBBTRFSUTIL_ALIAS(btrfs_util_delete_subvolume);
 
 PUBLIC enum btrfs_util_error btrfs_util_delete_subvolume_fd(int parent_fd,
 							    const char *name,
@@ -1291,6 +1347,28 @@ PUBLIC enum btrfs_util_error btrfs_util_delete_subvolume_fd(int parent_fd,
 
 	return BTRFS_UTIL_OK;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_delete_fd(int parent_fd,
+							    const char *name,
+							    int flags)
+LIBBTRFSUTIL_ALIAS(btrfs_util_delete_subvolume_fd);
+
+PUBLIC enum btrfs_util_error btrfs_util_delete_subvolume_by_id_fd(int parent_fd,
+							    uint64_t subvolid)
+{
+	struct btrfs_ioctl_vol_args_v2 args = {};
+	int ret;
+
+	args.flags = BTRFS_SUBVOL_SPEC_BY_ID;
+	args.subvolid = subvolid;
+
+	ret = ioctl(parent_fd, BTRFS_IOC_SNAP_DESTROY_V2, &args);
+	if (ret == -1)
+		return BTRFS_UTIL_ERROR_SNAP_DESTROY_FAILED;
+
+	return BTRFS_UTIL_OK;
+}
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_delete_by_id_fd(int fd, uint64_t subvolid)
+LIBBTRFSUTIL_ALIAS(btrfs_util_delete_subvolume_by_id_fd);
 
 PUBLIC void btrfs_util_destroy_subvolume_iterator(struct btrfs_util_subvolume_iterator *iter)
 {
@@ -1304,11 +1382,15 @@ PUBLIC void btrfs_util_destroy_subvolume_iterator(struct btrfs_util_subvolume_it
 		free(iter);
 	}
 }
+PUBLIC void btrfs_util_subvolume_iter_destroy(struct btrfs_util_subvolume_iterator *iter)
+LIBBTRFSUTIL_ALIAS(btrfs_util_destroy_subvolume_iterator);
 
 PUBLIC int btrfs_util_subvolume_iterator_fd(const struct btrfs_util_subvolume_iterator *iter)
 {
 	return iter->fd;
 }
+PUBLIC int btrfs_util_subvolume_iterator_get_fd(const struct btrfs_util_subvolume_iterator *iter)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_iterator_fd);
 
 static enum btrfs_util_error build_subvol_path(struct btrfs_util_subvolume_iterator *iter,
 					       const char *name, size_t name_len,
@@ -1361,6 +1443,9 @@ static enum btrfs_util_error build_subvol_path(struct btrfs_util_subvolume_itera
 	return BTRFS_UTIL_OK;
 }
 
+/*
+ * The @ref could be from raw search buffer, do not assume any alignment
+ */
 static enum btrfs_util_error build_subvol_path_privileged(struct btrfs_util_subvolume_iterator *iter,
 							  const struct btrfs_ioctl_search_header *header,
 							  const struct btrfs_root_ref *ref,
@@ -1368,8 +1453,8 @@ static enum btrfs_util_error build_subvol_path_privileged(struct btrfs_util_subv
 							  size_t *path_len_ret)
 {
 	struct btrfs_ioctl_ino_lookup_args lookup = {
-		.treeid = header->objectid,
-		.objectid = le64_to_cpu(ref->dirid),
+		.treeid = btrfs_search_header_objectid(header),
+		.objectid = get_unaligned_le64(&ref->dirid),
 	};
 	int ret;
 
@@ -1377,7 +1462,7 @@ static enum btrfs_util_error build_subvol_path_privileged(struct btrfs_util_subv
 	if (ret == -1)
 		return BTRFS_UTIL_ERROR_INO_LOOKUP_FAILED;
 
-	return build_subvol_path(iter, name, le16_to_cpu(ref->name_len),
+	return build_subvol_path(iter, name, get_unaligned_le16(&ref->name_len),
 				 lookup.name, strlen(lookup.name),
 				 path_len_ret);
 }
@@ -1444,21 +1529,30 @@ static enum btrfs_util_error subvolume_iterator_next_tree_search(struct btrfs_ut
 		header = (struct btrfs_ioctl_search_header *)(top->search.buf + top->buf_off);
 
 		top->items_pos++;
-		top->buf_off += sizeof(*header) + header->len;
-		top->search.key.min_offset = header->offset + 1;
+		top->buf_off += sizeof(*header) + btrfs_search_header_len(header);
+		top->search.key.min_offset = btrfs_search_header_offset(header) + 1;
 
 		/* This shouldn't happen, but handle it just in case. */
-		if (header->type != BTRFS_ROOT_REF_KEY)
+		if (btrfs_search_header_type(header) != BTRFS_ROOT_REF_KEY)
 			continue;
 
 		ref = (struct btrfs_root_ref *)(header + 1);
 		name = (const char *)(ref + 1);
 		err = build_subvol_path_privileged(iter, header, ref, name,
 						   &path_len);
-		if (err)
+		if (err) {
+			/*
+			 * If the subvolume's parent directory doesn't exist,
+			 * then the subvolume was either moved or deleted. Skip
+			 * it.
+			 */
+			if (errno == ENOENT)
+				continue;
 			return err;
+		}
 
-		err = append_to_search_stack(iter, header->offset, path_len);
+		err = append_to_search_stack(iter,
+				btrfs_search_header_offset(header), path_len);
 		if (err)
 			return err;
 
@@ -1524,8 +1618,12 @@ static enum btrfs_util_error subvolume_iterator_next_unprivileged(struct btrfs_u
 		err = build_subvol_path_unprivileged(iter, treeid, dirid,
 						     &path_len);
 		if (err) {
-			/* Skip the subvolume if we can't access it. */
-			if (errno == EACCES)
+			/*
+			 * If the subvolume's parent directory doesn't exist,
+			 * then the subvolume was either moved or deleted. Skip
+			 * it. Also skip it if we can't access it.
+			 */
+			if (errno == ENOENT || errno == EACCES)
 				continue;
 			return err;
 		}
@@ -1574,6 +1672,10 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_iterator_next(struct btrfs_uti
 							    id_ret);
 	}
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_iter_next(struct btrfs_util_subvolume_iterator *iter,
+							     char **path_ret,
+							     uint64_t *id_ret)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_iterator_next);
 
 PUBLIC enum btrfs_util_error btrfs_util_subvolume_iterator_next_info(struct btrfs_util_subvolume_iterator *iter,
 								     char **path_ret,
@@ -1591,6 +1693,10 @@ PUBLIC enum btrfs_util_error btrfs_util_subvolume_iterator_next_info(struct btrf
 	else
 		return btrfs_util_subvolume_info_fd(iter->cur_fd, 0, subvol);
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_iter_next_info(struct btrfs_util_subvolume_iterator *iter,
+								 char **path_ret,
+								 struct btrfs_util_subvolume_info *subvol)
+LIBBTRFSUTIL_ALIAS(btrfs_util_subvolume_iterator_next_info);
 
 PUBLIC enum btrfs_util_error btrfs_util_deleted_subvolumes(const char *path,
 							   uint64_t **ids,
@@ -1607,6 +1713,8 @@ PUBLIC enum btrfs_util_error btrfs_util_deleted_subvolumes(const char *path,
 	SAVE_ERRNO_AND_CLOSE(fd);
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_list_deleted(const char *path, uint64_t **ids, size_t *n)
+LIBBTRFSUTIL_ALIAS(btrfs_util_deleted_subvolumes);
 
 PUBLIC enum btrfs_util_error btrfs_util_deleted_subvolumes_fd(int fd,
 							      uint64_t **ids,
@@ -1635,6 +1743,7 @@ PUBLIC enum btrfs_util_error btrfs_util_deleted_subvolumes_fd(int fd,
 	*n = 0;
 	for (;;) {
 		const struct btrfs_ioctl_search_header *header;
+		struct btrfs_util_subvolume_info subvol;
 
 		if (items_pos >= search.key.nr_items) {
 			search.key.nr_items = 4096;
@@ -1656,7 +1765,7 @@ PUBLIC enum btrfs_util_error btrfs_util_deleted_subvolumes_fd(int fd,
 		 * The orphan item might be for a free space cache inode, so
 		 * check if there's a matching root item.
 		 */
-		err = btrfs_util_subvolume_info_fd(fd, header->offset, NULL);
+		err = btrfs_util_subvolume_info_fd(fd, btrfs_search_header_offset(header), &subvol);
 		if (!err) {
 			if (*n >= capacity) {
 				size_t new_capacity;
@@ -1673,14 +1782,14 @@ PUBLIC enum btrfs_util_error btrfs_util_deleted_subvolumes_fd(int fd,
 				*ids = new_ids;
 				capacity = new_capacity;
 			}
-			(*ids)[(*n)++] = header->offset;
+			(*ids)[(*n)++] = btrfs_search_header_offset(header);
 		} else if (err != BTRFS_UTIL_ERROR_SUBVOLUME_NOT_FOUND) {
 			goto out;
 		}
 
 		items_pos++;
-		buf_off += sizeof(*header) + header->len;
-		search.key.min_offset = header->offset + 1;
+		buf_off += sizeof(*header) + btrfs_search_header_len(header);
+		search.key.min_offset = btrfs_search_header_offset(header) + 1;
 	}
 
 	err = BTRFS_UTIL_OK;
@@ -1692,3 +1801,5 @@ out:
 	}
 	return err;
 }
+PUBLIC enum btrfs_util_error btrfs_util_subvolume_list_deleted_fd(int fd, uint64_t **ids, size_t *n)
+LIBBTRFSUTIL_ALIAS(btrfs_util_deleted_subvolumes_fd);
