@@ -2,7 +2,7 @@
 #
 # Verify that subvolume sync waits until the subvolume is cleaned
 
-source "$TEST_TOP/common"
+source "$TEST_TOP/common" || exit
 
 check_prereq mkfs.btrfs
 check_prereq btrfs
@@ -10,18 +10,19 @@ check_prereq btrfs
 setup_root_helper
 
 prepare_test_dev
-run_check "$TOP/mkfs.btrfs" -f "$TEST_DEV"
+
+run_check_mkfs_test_dev
 run_check_mount_test_dev
 run_check $SUDO_HELPER chmod a+rw "$TEST_MNT"
 
 cd "$TEST_MNT"
 
 for i in `seq 5`; do
-	run_check dd if=/dev/zero of=file$i bs=1M count=10
+	run_check dd if=/dev/zero of="file$i" bs=1M count=10
 done
 
 for sn in `seq 4`;do
-	run_check $SUDO_HELPER "$TOP/btrfs" subvolume snapshot . snap$sn
+	run_check $SUDO_HELPER "$TOP/btrfs" subvolume snapshot . "snap$sn"
 	for i in `seq 10`; do
 		run_check dd if=/dev/zero of="snap$sn/file$i" bs=1M count=10
 	done
@@ -30,7 +31,7 @@ done
 run_check $SUDO_HELPER "$TOP/btrfs" subvolume list .
 run_check $SUDO_HELPER "$TOP/btrfs" subvolume list -d .
 
-idtodel=`run_check_stdout $SUDO_HELPER "$TOP/btrfs" inspect-internal rootid snap3`
+idtodel=$(_get_subvolid snap3)
 
 # delete, sync after some time
 run_check $SUDO_HELPER "$TOP/btrfs" subvolume delete -c snap3
